/*
 * asm1825 - Assembler for PIC16F1825 and others.
 *
 * $Id: asm1825.c,v 1.3 2021-02-26 13:31:03-05 andrew Exp andrew $
 * 
 * Copyright (c) 2017-2020 Andrew Turnquist
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 * 
 */

#include <ctype.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <unistd.h>

#include "conf.h"
#include "lc.h"
#include "util.h"
#include "memmap.h"
#include "macro.h"
#include "source.h"
#include "opcodes.h"
#include "pragmas.h"
#include "processor.h"
#include "psect.h"
#include "symtab.h"
#include "eval.h"
#include "config.h"

const char rcsid[] = "$Id: asm1825.c,v 1.3 2021-02-26 13:31:03-05 andrew Exp andrew $";

// mmhandle chipmem;
char fnroot[MAXPATHLEN];


/* Source file stack */
struct sfstack {
  struct sfstack *next;
  FILE *fp;
  char *fn;
  int lineno;
};

struct sfstack *source_stack = NULL;


void strreplace(char *s, const char *m, const char *r)
{
  char *loc;
  int shift;
  int src_shift = 0;
  int i;

  MESSAGE(MSG_VERBOSE, "strreplace:  in: %s in %s", m, s);
  for (loc = strstr(s, m); loc != NULL; loc = strstr(s, m) ) {
    shift = (strlen(r)-strlen(m));
    if (shift < 0) { shift = 0; src_shift = 1; }
    MESSAGE(MSG_DEBUG, "       s: %s", s);
    MESSAGE(MSG_DEBUG, "  loc(1): %s, shift=%d, src_shift=%d",
        loc, shift, src_shift);
    MESSAGE(MSG_DEBUG, "  loc(2): dest = '%s', src = '%s', n = %d",
		loc + shift, loc + src_shift, strlen(loc)+1);
    memmove((loc + shift), loc + src_shift, strlen(loc)+1);
    MESSAGE(MSG_DEBUG, "  loc(3): %s", loc);
    i = 0;
    while (r[i] != '\0') {
      loc[i] = r[i];
      i++;
    }
  }
  MESSAGE(MSG_VERBOSE, "strreplace: out: %s in %s", m, s);
}


int splitargs(char *argstr, char *argv[])
{
  char *args = argstr;	// Don't modify argstr?
  int argc = 0;

  while (*args != '\0') {
    argv[argc++] = args;
    while ((*args != ',') && (*args != '\0')) args++;
    if (*args == ',') *(args++) = '\0';
  }
  argv[argc] = NULL;
  return argc;
}


void
do_org(const char *o)
{
  int v;

  if ((v = evalString(o)) == EVAL_FAIL) {
    // Label?
    message(MSG_WARNING, "ORG parameter of '%s' not understood", o);
  }

  MESSAGE(MSG_INFO, "Setting LC to 0x%04x", v);
  lcSet(v);
  bsr = -1;	// Invalidate BSR tracking
}




void do_config(int word, const char *op)
{
  unsigned short cval;

  cval = symResolveLiteral(op);
  MESSAGE(MSG_INFO, "Writing Config 0x%04x = 0x%04x",
      proc.confbase + 0x0006 + word, cval);
  mmWrite(chipmem, proc.confbase + 0x0006 + word, cval);
}


void do_scope(const int scope, char *op)
{
  int argc;
  int i;
  char *argv[16];

  argc = splitargs(op, argv);
  i = 0;
  while (i < argc) {
    symSetScope(argv[i], scope);
    i++;
  }
}


void do_equ(const char *label, const char *op, const int set)
{
  int i;
  int f;

  // sscanf(op, "%i", &i);
  i = evalString(op);
  MESSAGE(MSG_DEBUG, "Getting flags for label '%s'", label);
  if ((f = symFlags(label)) != EVAL_FAIL) {
    MESSAGE(MSG_VERBOSE, "Symbol flags = 0x%x", f);
    /* Symbol exists; okay for SET, !okay for EQU */
    if ((f & SF_MOD) || !(f & SF_DECLARED)) {
      /* Modifiable symbol.  Okay for SET */
      /* Should we still modify if new value is from an EQU? */
      symModify(label, i, SF_ABS | (SF_MOD * set));
      return;
    } else {
      MESSAGE(MSG_ERROR, "Attempt to redefine read-only symbol %s", label);
      exit(1);
    }
  }
  symAdd(label, i, SF_ABS | (SF_MOD * set));
}


FILE * do_include(FILE *sfp, const char *op)
{
  struct sfstack *sp;
  FILE *fp;
  char pstr[MAXPATHLEN];

  // First, push the current file pointer onto the stack.
  sp = (struct sfstack *)malloc(sizeof(struct sfstack));
  sp->fp = sfp;
  sp->fn = srcname;
  sp->lineno = srcline;
  sp->next = source_stack;
  source_stack = sp;

  // Next, open the new file.
  // Try current directory first.
  if ((fp = fopen(op, "r")) == NULL) {
    if (op[0] != '/') {
      snprintf(pstr, sizeof(pstr), "%s/%s", ASMINCLUDE_PATH, op);
      if ((fp = fopen(pstr, "r")) == NULL) {
        message(MSG_ERROR, "include file '%s' not found.\n", op);
        exit(1);
      }
    } else {
      message(MSG_ERROR, "include file '%s' not found.\n", op);
      exit(1);
    }
  }

  srcname = strdup(op);
  srcline = 0;
  // And return the new file pointer.
  return fp;
}


int valid_label_char(const char c)
{
  if (isalnum(c)) return 1;
  if (c == '_') return 1;
  return 0;
}


void macro_expand_labels(char *op, int oplen, const char *mname, const int inv)
{
  char *p = op;
  char *ss;
  char *se;
  char label[64];
  char olabel[64];
  int diff;

  MESSAGE(MSG_VERBOSE, "Performing macro label expansion on '%s'", op);
  while (((p-op) < oplen) && *p) {
    if (*p == '@') {
      /* Potential substitution */
      MESSAGE(MSG_VERBOSE, "  Potential substitution at %d", (p-op));
      /* Substition only valid at start of label (no preceeding label char */
      if ((p == op) || !valid_label_char(*(p-1))) {
        MESSAGE(MSG_VERBOSE, "  Attempting substitution");
        /* Should be safe to substitute */
        ss = p++;
        while ((((p-op) < oplen) && *p) && valid_label_char(*p)) p++;
        se = p;
        /* Now we need to remove [ss--(se-1)] from the string and
         * replace it with the localized version.
         */
        MESSAGE(MSG_VERBOSE, "  ss = %d, se = %d", (ss - op), (se - op));
        if ((se - ss) > 1) { /* More than just '@' */
          strncpy(olabel, ss, (se-ss));
          olabel[(se-ss)] = '\0';
          sprintf(label, "_@%s_%s%03d", mname, olabel, inv);
          diff = strlen(label) - strlen(olabel);
          MESSAGE(MSG_INFO, "'%s' --> '%s', diff = %d", olabel, label, diff);
          memmove(se+diff, se, (oplen - ((se-p)+diff)));
          strncpy(ss, label, strlen(label));
          p = se+diff-1;  /* -1 to accomodate p++ at end of while loop */
        }  /* if label passes tests */
      }  /* if start of label */
    }  /* if '@' */
    p++;
  }  /* while */
}


// Process (insert) a defined macro.  May be called recursively
// from itself for macro calls within a macro.
//
void do_macro(const char *mname, char *params, const char *cmt)
{
  StateList *sl;
  char label[LABEL_LEN];
  char labeltemp[LABEL_LEN];	/* Label template */
  char opcode[32];
  char operands[128];
  char comment[256];
  char *argv[10];
  int argc;
  int inv;
  int i;
  char mark[4];

  if ((sl = macLookup(mname)) == NULL) {
    MESSAGE(MSG_ERROR, "Macro '%s' not defined!", mname);
    exit(1);
  }
  inv = macInvocations(mname, 1);
  // fprintf(stderr, "Calling splitargs(\"%s\", ...)\n", params);
  argc = splitargs(params, argv);
  MESSAGE(MSG_INFO, "Macro %s called with %d arguments", mname, argc);
  // for (i=0; i<argc; i++) fprintf(stderr, "  %d: %s\n", i, argv[i]);
  if (lstp && cmt && (cmt[0] != '\0')) fprintf(lstp, "%04x\t\t\t;%s\n", psAddrBase(curpsect->pnum) + lcGet(), cmt);
  while (sl != NULL) {
    stUse(&sl, labeltemp, opcode, operands, comment);

    /* If there's a label in the macro, replace it in the form
     * of _@<macro name>_<label><xxx>, where <xxx> is the invocation count.
     * (A label beginning with _@ should not appear in input source.)
     */
    if (labeltemp[0] != '\0') {
        sprintf(label, "_@%s_%s%03d", mname, labeltemp, inv);
        MESSAGE(MSG_INFO, "  Substituting label %s --> %s\n", labeltemp, label);
    } else label[0] = '\0';
    // if (label[0] == '@') sprintf(&(label[strlen(label)]), "%d", inv);

    for (i=0; i<argc; i++) {
      sprintf(mark, "$%d", i+1);
      strreplace(operands, mark, argv[i]);
      // This isn't perfect; works only when the local label is the
      // only operand.  But it should make local labels work right in
      // most cases.
      // It seems this should really be performed on argv[i], but
      // that will require expanding (replacing) argv[i]'s string space.
      // if (operands[0] == '@') {
      //   strncpy(labeltemp, operands, sizeof(labeltemp));
      //   sprintf(operands, "_@%s_%s%03d", mname, labeltemp, inv);
      //   MESSAGE(MSG_INFO, "  Substituting operands %s --> %s\n", labeltemp, operands);
      // }
    }
    macro_expand_labels(operands, sizeof(operands), mname, inv);
    if (pass == 1)
        MESSAGE(MSG_INFO, "%04x\t%s\t%s %s\t%s", lcGet(), label, opcode, operands, comment);

    // Macro local labels and parameter substitions done. Now process the line.
    //

    // Add any label to the symbol list.
    //
    if ((isalnum(label[0])) || (label[0] == '_') || (label[0] == '@')) {
      if (pass == 0) {
        // Pass 1
        // Don't add the symbol if the opcode is the EQU directive
        if (strcmp(opcode, "equ") && strcmp(opcode, "set")) {
          MESSAGE(MSG_INFO, "Adding label %s = 0x%04x to symbol table", label, lcGet());
          symAdd(label, lcGet(), 0);
          bsr = -1;  // Invalidate BSR tracking at label
        }
      } else {
        // Pass 2
        // This should never be true unless there is a bug in the assember
        //
        if ((strcmp(opcode, "equ") && strcmp(opcode, "set")) && 
              (symLookup(label) != (psAddrBase(curpsect->pnum) + lcGet()))) {
          message(MSG_CRITICAL, "Second pass value mismatch on label '%s'",
              label);
          message(MSG_CRITICAL, "  symLookup(label) = 0x%04x, lcGet() = 0x%04x", symLookup(label), lcGet());
        }
        if (lstp) fprintf(lstp, "%04x\t\t%s:\n", psAddrBase(curpsect->pnum) + lcGet(), label);
        bsr = -1;  // Invalidate BSR tracking on label
      }  // if (pass)
    }

    // Process the opcode, if one exists
    //
    if (opcode[0] != '\0') {
      // fprintf(stderr, "DEBUG: processing opcode '%s'\n", opcode);
      try_opcode(opcode, operands, comment);
      if (!strcmp(opcode, "banksel")) do_banksel(operands, 1);
      if (macInvocations(opcode, 0) >= 0) do_macro(opcode, operands, comment);
    }

  }
}


FILE * pop_source(FILE *sfp)
{
  struct sfstack *sp;
  FILE *fp;

  // Close the current file.
  fclose(sfp);

  // If the stack is empty, return NULL. This means the main
  // source file has been closed.
  //
  if (!source_stack) return NULL;

  // Otherwise pop the last file pointer off the stack.
  fp = source_stack->fp;
  if (srcname) free(srcname);
  srcname = source_stack->fn;
  srcline = source_stack->lineno;
  sp = source_stack;
  source_stack = source_stack->next;
  free(sp);

  // And return the pointer.
  return fp;
}


void pass1(const char *sf)
{
  FILE *sfp;
  char ibuf[512];
  char label[LABEL_LEN];
  char opcode[32];
  char operands[128];
  char comment[256];
  char macname[32];
  StateList **msl = NULL;

  srcname = strdup(sf);
  // And make sure location counter is reset to start.
  if (flatmodel) lcSet(0); else psResetAll();

  pass = 0;
  macname[0] = '\0';	// Not in a macro define at start

  message(MSG_NOTICE, "Beginning pass 1");

  if ((sfp = fopen(sf, "r")) == NULL) {
    message(MSG_ERROR, "Error opening source file.");
    exit(1);
  }
  srcline = 0;

  symInit();

  do {
    while (fgets(ibuf, sizeof(ibuf)-1, sfp) != NULL) {
      srcline++;
      if (!srcParseLine(ibuf, label, opcode, operands, comment)) continue;
      lowercase(opcode);
      if (macname[0] != '\0') {
        // Reading and storing macro
        if (!strcmp(opcode, "endm")) {
          // fprintf(stderr, "Ending macro definition '%s'\n", macname);
          macname[0] = '\0';
        } else {
          // Store this line in the macro
          stAdd(msl, label, opcode, operands, comment);
        }
      } else {
        if ((isalnum(label[0])) || (label[0] == '_') || (label[0] == '@')) {
          // Don't add the symbol if the opcode is the EQU directive
          if (strcmp(opcode, "equ") && strcmp(opcode, "set")) {
            symAdd(label, lcGet(), 0);
            bsr = -1;  // Invalidate BSR tracking at label
          }
        }

        if (opcode[0] != '\0') {
          MESSAGE(MSG_DEBUG, "processing opcode '%s'", opcode);
          try_opcode(opcode, operands, comment);

          // Right now we're just invalidating BSR tracking if the
          //   opcode was used.  We could read the operand
          //   and update 'bsr' accordingly.
          // movlb checking now takes place in opcodes.c
          // If everything checks out, this can be removed here.
          // if (!strcmp(opcode, "movlb")) bsr = -1;

          if (!strcmp(opcode, "end")) {
            MESSAGE(MSG_INFO, "END statement reached.");
            // Close all input files
            while ((sfp = pop_source(sfp)));
            // Break out of the loops
            return;
          }
          if (!strcmp(opcode, "macro")) {
            MESSAGE(MSG_INFO, "Defining macro '%s'", operands);
            strncpy(macname, operands, 32);
            msl = macAdd(macname, 0);
          }
          if (!strcmp(opcode, "banksel")) do_banksel(operands, 1);
          if (!strcmp(opcode, "endm")) macname[0] = '\0';
          if (!strcmp(opcode, "equ")) do_equ(label, operands, 0);
          if (!strcmp(opcode, "include")) sfp = do_include(sfp, operands);
          if (!strcmp(opcode, "org")) do_org(operands);
          if (!strcmp(opcode, "pragma")) pragDoDirective(operands);
          if (!strcmp(opcode, "global")) do_scope(SF_SCOPE_GLOBAL, operands);
          if (!strcmp(opcode, "file")) do_scope(SF_SCOPE_FILE, operands);
          if (!strcmp(opcode, "processor")) procDoDirective(operands);
          if (!strcmp(opcode, "psect")) psDoDirective(operands);
          if (!strcmp(opcode, "set")) do_equ(label, operands, 1);
          if (macInvocations(opcode, 0) >= 0) do_macro(opcode, operands, comment);
        }  // opcode not empty
      }  // Not in macro
    }
  } while ((sfp = pop_source(sfp)));

  // The last sfp will have been closed by pop_source
}


void pass2(const char *sf)
{
  FILE *sfp;
  char ibuf[512];
  char label[LABEL_LEN];
  char opcode[32];
  char operands[128];
  char comment[256];
  int voc;		/* Valid opcode flag */
  int done = 0;
  int inmacro = 0;

  srcname = strdup(sf);
  // Make sure location counter is reset to start.
  srcline = 0;
  if (flatmodel) lcSet(0); else psResetAll();
  psDoDirective(DEFAULT_PSECT);
  pass = 1;
  message(MSG_NOTICE, "Beginning pass 2");
  macReset();

  sprintf(ibuf, "%s.lst", fnroot);
  if ((lstp = fopen(ibuf, "w")) == NULL) {
    message(MSG_WARNING, "Could not open list file for writing");
  }

  if ((sfp = fopen(sf, "r")) == NULL) {
    message(MSG_ERROR, "Error opening source file.");
    exit(1);
  }

  do {
    while (fgets(ibuf, sizeof(ibuf)-1, sfp) != NULL) {
      srcline++;
      if (!srcParseLine(ibuf, label, opcode, operands, comment)) continue;
      lowercase(opcode);

      // Macro definitions are skipped in second pass
      //
      if (inmacro) {
        if (!strcmp(opcode, "endm")) { voc=1; inmacro = 0; }
        continue;
      }

      if ((isalnum(label[0])) || (label[0] == '_') || (label[0] == '@')) {

        // This should never be true unless there is a bug in the assember
        //
        if (strcmp(opcode, "equ") && strcmp(opcode, "set")) {
          // if (symLookup(label) != (lcGet())) {
          if (symLookup(label) != (psAddrBase(curpsect->pnum) + lcGet())) {
            message(MSG_CRITICAL, "Second pass value mismatch on label '%s'. Assembler error?",
                label);
            message(MSG_CRITICAL, "  psect %d, sym value 0x%04x, lc 0x%04x, AddrBase 0x%04x",
                curpsect->pnum, symLookup(label), lcGet(), psAddrBase(curpsect->pnum));
            message(MSG_CRITICAL, "  curpsect->start = 0x%04x (%d), curpsect->lc = 0x%04x", curpsect->start, curpsect->start, curpsect->lc);
          }
          if (lstp) fprintf(lstp, "%04x\t\t%s:\n", psAddrBase(curpsect->pnum) + lcGet(), label);
          bsr = -1;  // Invalidate BSR tracking on label
        }
      }

      if (opcode[0] != '\0') {
      
        // fprintf(stderr, "Processing opcode: '%s'\n", opcode);

        if (try_opcode(opcode, operands, comment) < 0) {
          voc = 0;
          if (!strcmp(opcode, "end")) {
            MESSAGE(MSG_INFO, "%04x END statement reached.", lcGet());
            // Close all input files
            while ((sfp = pop_source(sfp)));
            // Break out of the loops
            done = 1;
            break;
          }
          if (!strcmp(opcode, "equ")) voc = 1;
          if (!strcmp(opcode, "include")) {
            voc = 1;
            sfp = do_include(sfp, operands);
          }
          if (!strcmp(opcode, "banksel")) { voc=1; do_banksel(operands, 1); }
          if (!strcmp(opcode, "macro")) { voc=1; inmacro = 1; }
          if (!strcmp(opcode, "endm")) { voc=1; inmacro = 0; }
          if (!strcmp(opcode, "org")) { voc=1; do_org(operands); }
          if (!strcmp(opcode, "pragma")) { voc=1; pragDoDirective(operands); }
          if (!strcmp(opcode, "processor"))
              { voc=1; procDoDirective(operands); }
          if (!strcmp(opcode, "psect")) { voc=1; psDoDirective(operands); }
          if (!strcmp(opcode, "set")) voc = 1;
          if (!strcmp(opcode, "global"))
              { voc=1; do_scope(SF_SCOPE_GLOBAL, operands); }
          if (!strcmp(opcode, "file"))
              { voc=1; do_scope(SF_SCOPE_FILE, operands); }
          if (!strcmp(opcode, "__config"))  { voc=1; do_config(1, operands); }
          if (!strcmp(opcode, "__config1")) { voc=1; do_config(1, operands); }
          if (!strcmp(opcode, "__config2")) { voc=1; do_config(2, operands); }
          if (!strcmp(opcode, "__config3")) { voc=1; do_config(3, operands); }
          if (!strcmp(opcode, "__config4")) { voc=1; do_config(4, operands); }
          if (macInvocations(opcode, 0) >= 0) {
            voc=1;
            do_macro(opcode, operands, comment);
          }
          if (!voc) message(MSG_WARNING, "%04x: Invalid opcode or directive '%s'", lcGet(), opcode);
        } else {
          // Opcode found. Normally do nothing.
        }

      }

    }
  } while (!done && (sfp = pop_source(sfp)));

  // Last sfp will have been closed by pop_source.

  fclose(lstp);
}


/*
 * Sort all psects and assign memory start locations.  To be
 * called after pass1(), so pass2() knows where to put code segments
 * (and EEPROM data).
 */

void sortpsects()
{
}



void
getroot(char *r, const char *f)
{
  int i;

  strncpy(r, f, 1023);
  i = strlen(r);
  while ((r[i] != '.') && (i > 0)) i--;
  if (r[i] == '.') r[i] = '\0';
}


int main(int argc, char *argv[])
{
  int opt;
  FILE *symp;
  char *src;
  int banks;
  char ibuf[512];

  pass = 0;
  srcline = 0;
  srcname = "";
  procInit();
  while ((opt = getopt(argc, argv, "flpqvV")) != -1) {
    switch (opt) {
      case 'f': flatmodel = 1; break;	/* flat */
      case 'p': flatmodel = 0; break;	/* psects */
      case 'l': smartbsr = 0;  break;	/* Literal BANKSEL */
      case 'q': msg_level--;   break;
      case 'v': msg_level++;   break;
      case 'V':
        fprintf(stderr, "asm1825 version %s\n", VERSION);
      default:
        fprintf(stderr, "Usage: %s [ -flpqv ] <source>\n", argv[0]);
        exit(1);
    }
  }
  if (argv[optind] == NULL) {
    fprintf(stderr, "Usage: %s [ -fpqv ] <source>\n", argv[0]);
    exit(1);
  }

  src = argv[optind];

  chipmem = mmInit(8192, 0x0b, 256, 0x8000);

  /* Calculate number of banks */
  banks = (proc.ramsize - 16) / 80;
  /* Add 1 if there's a partial bank at end */
  if (((banks * 80) + 16) != proc.ramsize) banks++;
  psInit(banks);

  message(MSG_NOTICE, "%s %s starting", PACKAGE, VERSION);

  getroot(fnroot, src);
  MESSAGE(MSG_INFO, "Using root name: '%s'\n", fnroot);

  ocitInit();
  macInit();


  pass1(src);
  // macDump(stdout);

  // sortpsects();
  psAssign();

  pass2(src);


  fprintf(stderr, "\n");
  sprintf(ibuf, "%s.hex", fnroot);
  mmWriteHexfile(chipmem, ibuf);

  sprintf(ibuf, "%s.sym", fnroot);
  if ((symp = fopen(ibuf, "w")) != NULL) {
    symDump(symp);
    fclose(symp);
  }

  sprintf(ibuf, "%s.psl", fnroot);
  if ((symp = fopen(ibuf, "w")) != NULL) {
    psDump(symp);
    fclose(symp);
  }


  symDestroy();
  macDestroy();
  ocitDestroy();

  mmFree(chipmem);

  exit(0);
}


