/*
 * asm1825 - Assembler for PIC16F1825 and others.
 *
 * $Id: opcodes.c,v 1.1 2021-02-26 13:33:35-05 andrew Exp andrew $
 * 
 * Copyright (c) 2017 Andrew Turnquist
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 * 
 */

#include <ctype.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "lc.h"
#include "memmap.h"
#include "processor.h"
#include "psect.h"
#include "symtab.h"
#include "eval.h"
#include "util.h"
#include "opcodes.h"



int bsr = -1;
int pass;
FILE *lstp;

extern mmhandle chipmem;

int memmodel = MM_LARGE;

#define OC_FORWARD	(1 << 0)	/* Opcode had a forward reference */
					/* in pass 1                      */

struct oc_info {
  struct oc_info *next;
  int addr;
  int psect;
  unsigned int flags;
};

#define INFOTAB_SIZE	32

struct oc_info *oc_infotab[INFOTAB_SIZE];


void ocitInit(void)
{
  int i;

  for(i=0; i<INFOTAB_SIZE; i++) {
    oc_infotab[i] = NULL;
  }
}


void ocitDestroy(void)
{
  int i;
  struct oc_info *op;

  for (i=0; i<INFOTAB_SIZE; i++) {
    while (oc_infotab[i] != NULL) {
      op = (oc_infotab[i])->next;
      free(oc_infotab[i]);
      oc_infotab[i] = op;
    }
  }
}


// TODO: Add parameter to specify a psect (0 = any psect)
struct oc_info *ocitFindEntry(const int loc, const int ps)
{
  int i;
  struct oc_info *op;

  i = (loc % INFOTAB_SIZE);
  op = oc_infotab[i];
  while (op) {
    if ( (op->addr == loc) && ((ps == 0) || (op->psect == ps))) return op;
    op = op->next;
  }
  return NULL;
}


// TODO: Accept psect parameter and pass as qualifier to ocitFindEntry
void ocitSetFlag(const int loc, const int ps, unsigned int f)
{
  int i;
  struct oc_info *oent;

  // fprintf(stderr, "%04x ocitSetFlag: Setting flags %04x at address %04x\n", lcGet(), f, loc);
  if ((oent = ocitFindEntry(loc, 0)) != NULL) {
    oent->flags |= f;
  } else {
    i = (loc % INFOTAB_SIZE);
    oent = (struct oc_info *)malloc(sizeof(struct oc_info));
    oent->addr = loc;
    oent->psect = curpsect->pnum;
    oent->flags = f;
    oent->next = oc_infotab[i];
    oc_infotab[i] = oent;
  }
}


/*
 *  Checks if flag(s) 'f' is/are set for location 'loc'. Returns
 *  bitwise AND of flags and 'f'. Return value is 'f' if all flags
 *  set, >0 if any flag is set, and 0 if no flags are set.
 */
unsigned int ocitCheckFlag(const int loc, unsigned int f)
{
  int i;
  struct oc_info *op;

  i = (loc % INFOTAB_SIZE);
  op = oc_infotab[i];
  while (op) {
    if (op->addr == loc) return (op->flags & f);
    op = op->next;
  }
  return 0;
}



void split_operands(char *op, char **op1, char **op2)
{
  char *temp_op;
  char quoting = '\0';

  // fprintf(stderr, "%04x DEBUG: Entering split_operands: '%s'\n", lc, op);
  *op1 = op;
  temp_op = op;
  while (*temp_op) {
    if (quoting && (*temp_op == quoting)) {
      quoting = '\0';
      temp_op++;
      continue;
    }
    if ((*temp_op == '\'') || (*temp_op == '"')) {
      quoting = *temp_op;
      temp_op++;
      continue;
    }
    if (*temp_op == ',') {
      *temp_op = '\0';
      temp_op++;
      while (*temp_op != '\0') {
        if ((*temp_op != ' ') && (*temp_op != '\t')) break;
        temp_op++;
      }
      *op2 = temp_op;
      return;
    }
    temp_op++;
  }
  message(MSG_ERROR, "split_operands: Never found operand 2");
  *op2 = NULL;
}


int parse_fb(const char *op, unsigned short *f, unsigned short *b)
{
  char buf[128];
  char *fstr;
  char *bstr;
  int ff;	/* Full value of fstr symbol (not shortened) */

  strncpy(buf,op, sizeof(buf)-1);
  split_operands(buf, &fstr, &bstr);
  // fprintf(stderr, "DEBUG: parse_fb: f = '%s', b = '%s'\n", fstr, bstr);

  if ((fstr[0] == '\0') || (bstr[0] == '\0')) return -1;
  ff = symResolveLiteral(fstr);
  *f = ff & 0x007f;
  // *f = symResolveLiteral(fstr) & 0x007f;
  *b = symResolveLiteral(bstr) & 0x0007;
  return ff;
}



int parse_fd(const char *op, unsigned short *f, unsigned short *d)
{
  char buf[128];
  char *fstr;
  char *dstr;
  int ff;	/* Full value of fstr symbol (not shortened) */

  strncpy(buf,op, sizeof(buf)-1);
  split_operands(buf, &fstr, &dstr);
  // fprintf(stderr, "DEBUG: parse_fd: f = '%s', d = '%s'\n", fstr, dstr);

  if ((fstr[0] == '\0') || (dstr[0] == '\0')) return -1;
  ff = symResolveLiteral(fstr);
  *f = ff & 0x007f;
  // *f = symResolveLiteral(fstr) & 0x007f;
  lowercase(dstr);
  *d = 0xffff;
  if ((dstr[0] == '0') || (dstr[0] == 'w')) *d = 0;
  if ((dstr[0] == '1') || (dstr[0] == 'f')) *d = 1;
  if (*d == 0xffff) {
    message(MSG_WARNING, "Invalid destination '%s'. Assuming W", dstr);
    *d = 0;
    return -1;
  }
  return ff;
}


int iscore(const unsigned short fl)
{
  if (fl >= 32) return 0;
  return proc.coremask & (1 << fl);
}


/***** Directives *****/

void do_banksel(const char *op, const int lit)
{
  unsigned short f = evalString(op);
  unsigned short fl = f & 0x007f;
  unsigned short cval = BANK(f);
  char buf[8];

  if (!smartbsr && !lit) {
    MESSAGE(MSG_INFO, "Skipping auto-generated BANKSEL %s", op);
    return;
  }

  /* Core registers and common RAM are bank independent */
  if (smartbsr && (iscore(fl) || (fl >= 0x70))) {
    if (lit) MESSAGE(MSG_INFO, "Ignoring redundant BANKSEL %s", op);
    return;
  }

  if (!smartbsr || ((bsr < 0) || (bsr != cval))) {
    if (proc.inset & IS_EXTENDED) {
      snprintf(buf, sizeof(buf), "0x%x", cval);
      if (lit) {
        try_opcode("movlb", buf, "; Generated by BANKSEL");
      } else {
        try_opcode("movlb", buf, "; Generated by BANKSEL (auto)");
      }
      // bsr = cval;
    } else {
      /* No movlb opcode; set bits in STATUS */
      snprintf(buf, sizeof(buf), "%s", cval ? "bsf" : "bcf");
      bsr = cval & 0x0007;
      if (lit) {
        try_opcode(buf, "0x03,5", "; Generated by BANKSEL");
      } else {
        try_opcode(buf, "0x03,5", "; Generated by BANKSEL (auto)");
      }
    }
  } else {
    if (!smartbsr) MESSAGE(MSG_INFO, "Ignoring redundant BANKSEL %s", op);
  }
}



/***** Byte-Oriented File Register Operations *****/

void do_addwf(const char *op)
{
  unsigned short f;
  unsigned short d;
  int fl;
  char fls[8];

  if ((fl = parse_fd(op, &f, &d)) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (BANK(fl) != bsr) {
    sprintf(fls, "0x%04x", fl);
    do_banksel(fls, 0);
  }

  if ((f == R_BSR) && (d == 1)) bsr = -1;
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x0700 | (d << 7) | f);
  lcInc();
}


void do_addwfc(const char *op)
{
  unsigned short f;
  unsigned short d;
  int fl;
  char fls[8];

  if ((fl = parse_fd(op, &f, &d)) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (!(proc.inset & IS_EXTENDED))
      MESSAGE(MSG_WARNING, "Invalid opcode for this chip");

  if (BANK(fl) != bsr) {
    sprintf(fls, "0x%04x", fl);
    do_banksel(fls, 0);
  }

  if ((f == R_BSR) && (d == 1)) bsr = -1;
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x3d00 | (d << 7) | f);
  lcInc();
}


void do_andwf(const char *op)
{
  unsigned short f;
  unsigned short d;
  int fl;
  char fls[8];

  if ((fl = parse_fd(op, &f, &d)) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (BANK(fl) != bsr) {
    sprintf(fls, "0x%04x", fl);
    do_banksel(fls, 0);
  }

  if ((f == R_BSR) && (d == 1)) bsr = -1;
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x0500 | (d << 7) | f);
  lcInc();
}


void do_asrf(const char *op)
{
  unsigned short f;
  unsigned short d;
  int fl;
  char fls[8];

  if ((fl = parse_fd(op, &f, &d)) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (!(proc.inset & IS_EXTENDED))
      MESSAGE(MSG_WARNING, "Invalid opcode for this chip");

  if (BANK(fl) != bsr) {
    sprintf(fls, "0x%04x", fl);
    do_banksel(fls, 0);
  }

  if ((f == R_BSR) && (d == 1)) bsr = -1;
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x3700 | (d << 7) | f);
  lcInc();
}


void do_lslf(const char *op)
{
  unsigned short f;
  unsigned short d;
  int fl;
  char fls[8];

  if ((fl = parse_fd(op, &f, &d)) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (!(proc.inset & IS_EXTENDED))
      MESSAGE(MSG_WARNING, "Invalid opcode for this chip");

  if (BANK(fl) != bsr) {
    sprintf(fls, "0x%04x", fl);
    do_banksel(fls, 0);
  }

  if ((f == R_BSR) && (d == 1)) bsr = -1;
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x3500 | (d << 7) | f);
  lcInc();
}


void do_lsrf(const char *op)
{
  unsigned short f;
  unsigned short d;
  int fl;
  char fls[8];

  if ((fl = parse_fd(op, &f, &d)) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (!(proc.inset & IS_EXTENDED))
      MESSAGE(MSG_WARNING, "Invalid opcode for this chip");

  if (BANK(fl) != bsr) {
    sprintf(fls, "0x%04x", fl);
    do_banksel(fls, 0);
  }

  if ((f == R_BSR) && (d == 1)) bsr = -1;
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x3600 | (d << 7) | f);
  lcInc();
}


void do_clrf(const char *op)
{
  int fl = symResolveLiteral(op);
  uint16_t f = fl & 0x7f;
  char fls[8];

  if (BANK(fl) != bsr) {
    sprintf(fls, "0x%04x", fl);
    do_banksel(fls, 0);
  }

  if (f == R_BSR) bsr = -1;
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x0180 | f);
  lcInc();
}


void do_clrw(const char *op)
{
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x0100);
  lcInc();
}


void do_comf(const char *op)
{
  unsigned short f;
  unsigned short d;
  int fl;
  char fls[8];

  if ((fl = parse_fd(op, &f, &d)) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (BANK(fl) != bsr) {
    sprintf(fls, "0x%04x", fl);
    do_banksel(fls, 0);
  }

  if ((f == R_BSR) && (d == 1)) bsr = -1;
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x0900 | (d << 7) | f);
  lcInc();
}


void do_decf(const char *op)
{
  unsigned short f;
  unsigned short d;
  int fl;
  char fls[8];

  if ((fl = parse_fd(op, &f, &d)) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (BANK(fl) != bsr) {
    sprintf(fls, "0x%04x", fl);
    do_banksel(fls, 0);
  }

  // Only decrement 'bsr' if f = BSR, 'bsr' is valid, and dest is BSR
  if ((f == R_BSR) && (bsr >= 0) && (d == 1)) {
    bsr--;
    if (bsr < 0) bsr = 0x1f;
  }
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x0300 | (d << 7) | f);
  lcInc();
}


void do_incf(const char *op)
{
  unsigned short f;
  unsigned short d;
  int fl;
  char fls[8];

  if ((fl = parse_fd(op, &f, &d)) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (BANK(fl) != bsr) {
    sprintf(fls, "0x%04x", fl);
    do_banksel(fls, 0);
  }

  // Only increment 'bsr' if f = BSR, 'bsr' is valid, and dest is BSR
  if ((f == R_BSR) && (bsr >= 0) && (d == 1)) {
    bsr++;
    bsr &= 0x1f;
  }
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x0a00 | (d << 7) | f);
  lcInc();
}


void do_iorwf(const char *op)
{
  unsigned short f;
  unsigned short d;
  int fl;
  char fls[8];

  if ((fl = parse_fd(op, &f, &d)) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (BANK(fl) != bsr) {
    sprintf(fls, "0x%04x", fl);
    do_banksel(fls, 0);
  }

  if ((f == R_BSR) && (d == 1)) bsr = -1;
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x0400 | (d << 7) | f);
  lcInc();
}


void do_movf(const char *op)
{
  unsigned short f;
  unsigned short d;
  int fl;
  char fls[8];

  if ((fl = parse_fd(op, &f, &d)) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (BANK(fl) != bsr) {
    sprintf(fls, "0x%04x", fl);
    do_banksel(fls, 0);
  }

  if (pass) mmWrite(chipmem, lcGetAbs(), 0x0800 | (d << 7) | f);
  lcInc();
}


void do_movwf(const char *op)
{
  int fl = symResolveLiteral(op);
  uint16_t f = (fl & 0x7f);
  char fls[8];

  if (BANK(fl) != bsr) {
    sprintf(fls, "0x%04x", fl);
    do_banksel(fls, 0);
  }

  if (f == R_BSR) bsr = -1;
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x0080 | f);
  lcInc();
}


void do_rlf(const char *op)
{
  unsigned short f;
  unsigned short d;
  int fl;
  char fls[8];

  if ((fl = parse_fd(op, &f, &d)) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (BANK(fl) != bsr) {
    sprintf(fls, "0x%04x", fl);
    do_banksel(fls, 0);
  }

  if ((f == R_BSR) && (d == 1)) bsr = -1;
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x0d00 | (d << 7) | f);
  lcInc();
}


void do_rrf(const char *op)
{
  unsigned short f;
  unsigned short d;
  int fl;
  char fls[8];

  if ((fl = parse_fd(op, &f, &d)) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (BANK(fl) != bsr) {
    sprintf(fls, "0x%04x", fl);
    do_banksel(fls, 0);
  }

  if ((f == R_BSR) && (d == 1)) bsr = -1;
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x0c00 | (d << 7) | f);
  lcInc();
}


void do_subwf(const char *op)
{
  unsigned short f;
  unsigned short d;
  int fl;
  char fls[8];

  if ((fl = parse_fd(op, &f, &d)) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (BANK(fl) != bsr) {
    sprintf(fls, "0x%04x", fl);
    do_banksel(fls, 0);
  }

  if ((f == R_BSR) && (d == 1)) bsr = -1;
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x0200 | (d << 7) | f);
  lcInc();
}


void do_subwfb(const char *op)
{
  unsigned short f;
  unsigned short d;
  int fl;
  char fls[8];

  if ((fl = parse_fd(op, &f, &d)) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (!(proc.inset & IS_EXTENDED))
      MESSAGE(MSG_WARNING, "Invalid opcode for this chip");

  if (BANK(fl) != bsr) {
    sprintf(fls, "0x%04x", fl);
    do_banksel(fls, 0);
  }

  if ((f == R_BSR) && (d == 1)) bsr = -1;
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x3b00 | (d << 7) | f);
  lcInc();
}


void do_swapf(const char *op)
{
  unsigned short f;
  unsigned short d;
  int fl;
  char fls[8];

  if ((fl = parse_fd(op, &f, &d)) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (BANK(fl) != bsr) {
    sprintf(fls, "0x%04x", fl);
    do_banksel(fls, 0);
  }

  if ((f == R_BSR) && (d == 1)) bsr = -1;
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x0e00 | (d << 7) | f);
  lcInc();
}


void do_xorwf(const char *op)
{
  unsigned short f;
  unsigned short d;
  int fl;
  char fls[8];

  if ((fl = parse_fd(op, &f, &d)) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (BANK(fl) != bsr) {
    sprintf(fls, "0x%04x", fl);
    do_banksel(fls, 0);
  }

  if ((f == R_BSR) && (d == 1)) bsr = -1;
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x0600 | (d << 7) | f);
  lcInc();
}



/***** Byte Oriented Skip Operations *****/

void do_decfsz(const char *op)
{
  unsigned short f;
  unsigned short d;
  int fl;
  char fls[8];

  if ((fl = parse_fd(op, &f, &d)) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (BANK(fl) != bsr) {
    MESSAGE(MSG_INFO, "fl = %d, bank(fl) = %d, bsr = %d", fl, BANK(fl), bsr);
    sprintf(fls, "0x%04x", fl);
    do_banksel(fls, 0);
  }

  // Only decrement 'bsr' if f = BSR, 'bsr' is valid, and dest is BSR
  if ((f == R_BSR) && (bsr >= 0) && (d == 1)) {
    bsr--;
    if (bsr < 0) bsr = 0x1f;
  }
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x0b00 | (d << 7) | f);
  lcInc();
}


void do_incfsz(const char *op)
{
  unsigned short f;
  unsigned short d;
  int fl;
  char fls[8];

  if ((fl = parse_fd(op, &f, &d)) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (BANK(fl) != bsr) {
    MESSAGE(MSG_INFO, "fl = %d, bank(fl) = %d, bsr = %d", fl, BANK(fl), bsr);
    sprintf(fls, "0x%04x", fl);
    do_banksel(fls, 0);
  }

  // Only increment 'bsr' if f = BSR, 'bsr' is valid, and dest is BSR
  if ((f == R_BSR) && (bsr >= 0) && (d == 1)) {
    bsr++;
    bsr &= 0x1f;
  }
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x0f00 | (d << 7) | f);
  lcInc();
}



/***** Bit-Oriented File Register Operations *****/

void do_bcf(const char *op)
{
  unsigned short f;
  unsigned short b;
  int fl;
  char fls[8];

  if ((fl = parse_fb(op, &f, &b)) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (BANK(fl) != bsr) {
    MESSAGE(MSG_INFO, "fl = %d, bank(fl) = %d, bsr = %d", fl, BANK(fl), bsr);
    sprintf(fls, "0x%04x", fl);
    do_banksel(fls, 0);
  }

  if ((f == R_BSR) && (bsr >= 0)) {
    int bm = 0x1f ^ (1 << b);
    bsr &= bm;
  }
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x1000 | (b << 7) | f);
  lcInc();
}


void do_bsf(const char *op)
{
  unsigned short f;
  unsigned short b;
  int fl;
  char fls[8];

  if ((fl = parse_fb(op, &f, &b)) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (BANK(fl) != bsr) {
    MESSAGE(MSG_INFO, "fl = %d, bank(fl) = %d, bsr = %d", fl, BANK(fl), bsr);
    sprintf(fls, "0x%04x", fl);
    do_banksel(fls, 0);
  }

  if ((f == R_BSR) && (bsr >= 0)) {
    int bm = 1 << b;
    bsr |= bm;
  }
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x1400 | (b << 7) | f);
  lcInc();
}



/***** Bit-Oriented Skip Operations *****/

void do_btfsc(const char *op)
{
  unsigned short f;
  unsigned short b;
  int fl;
  char fls[8];

  if ((fl = parse_fb(op, &f, &b)) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (BANK(fl) != bsr) {
    MESSAGE(MSG_INFO, "fl = %d, bank(fl) = %d, bsr = %d", fl, BANK(fl), bsr);
    sprintf(fls, "0x%04x", fl);
    do_banksel(fls, 0);
  }

  if (pass) mmWrite(chipmem, lcGetAbs(), 0x1800 | (b << 7) | f);
  lcInc();
}


void do_btfss(const char *op)
{
  unsigned short f;
  unsigned short b;
  int fl;
  char fls[8];

  if ((fl = parse_fb(op, &f, &b)) == -1) {
    fprintf(stderr, "ERROR: parse_fb failed to parse operands.\n");
    exit(1);
  }

  if (BANK(fl) != bsr) {
    MESSAGE(MSG_INFO, "fl = %d, bank(fl) = %d, bsr = %d", fl, BANK(fl), bsr);
    sprintf(fls, "0x%04x", fl);
    do_banksel(fls, 0);
  }

  if (pass) mmWrite(chipmem, lcGetAbs(), 0x1c00 | (b << 7) | f);
  lcInc();
}



/***** Literal Operations *****/

void do_addlw(const char *op)
{
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x3e00 | (symResolveLiteral(op) & 0x00ff));
  lcInc();
}


void do_andlw(const char *op)
{
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x3900 | (symResolveLiteral(op) & 0x00ff));
  lcInc();
}


void do_iorlw(const char *op)
{
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x3800 | (symResolveLiteral(op) & 0x00ff));
  lcInc();
}


void do_movlb(const char *op)
{
  if (!(proc.inset & IS_EXTENDED))
      MESSAGE(MSG_WARNING, "Invalid opcode for this chip");

  bsr = symResolveLiteral(op) & 0x001f;
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x0020 | bsr);
  lcInc();
}


void do_movlp(const char *op)
{
  if (!(proc.inset & IS_EXTENDED))
      MESSAGE(MSG_WARNING, "Invalid opcode for this chip");

  if (pass) mmWrite(chipmem, lcGetAbs(), 0x3180 | (symResolveLiteral(op) & 0x007f));
  lcInc();
}


void do_movlw(const char *op)
{
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x3000 | (symResolveLiteral(op) & 0x00ff));
  lcInc();
}


void do_sublw(const char *op)
{
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x3c00 | (symResolveLiteral(op) & 0x00ff));
  lcInc();
}


void do_xorlw(const char *op)
{
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x3a00 | (symResolveLiteral(op) & 0x00ff));
  lcInc();
}



/***** Control Operations *****/

void
do_bra(const char *op)
{
  int dst;

  if (!(proc.inset & IS_EXTENDED))
      MESSAGE(MSG_WARNING, "Invalid opcode for this chip");

  dst = symResolveAddr(op);
  message(MSG_INFO, "do_bra(): dst = 0x%04x", dst);
  if (dst == EVAL_FAIL) {
    if (!pass) {
      message(MSG_INFO, "%04x: Forward label", lcGet());
      ocitSetFlag(lcGet(), curpsect->pnum, OC_FORWARD);
    }
    if (pass) {
      message(MSG_ERROR, "%04x: Unresolved symbol: %s", lcGetAbs(), op);
      exit(1);
    }
  }
  if (pass) {
    dst = (dst - (lcGetAbs() + 1));
    message(MSG_INFO, "do_bra(): edited dst = 0x%04x", dst);

    if ((dst < -256) || (dst > 255)) {
      message(MSG_WARNING, "bra operand (%d) out of range. Code will be wrong.", dst);
    }

    mmWrite(chipmem, lcGetAbs(), 0x3200 | (shorten2s(dst, 9)));
  }
  lcInc();
}


void
do_brw(const char *op)
{
  if (!(proc.inset & IS_EXTENDED))
      MESSAGE(MSG_WARNING, "Invalid opcode for this chip");

  if (pass) mmWrite(chipmem, lcGetAbs(), 0x000b);
  lcInc();
}


void
do_call_near(const char *op)
{
  int dst;

  dst = symResolveAddr(op);
  if (dst == EVAL_FAIL) {
    if (!pass) {
      message(MSG_NOTICE, "Forward label; near call forced");
      ocitSetFlag(lcGet(), curpsect->pnum, OC_FORWARD);
    }
    if (pass) {
      message(MSG_ERROR, "Unresolved symbol: %s\n", op);
      exit(1);
    }
  }
  if (pass) {
    /* Pass 2 */
    if ((dst == EVAL_FAIL) || ((dst & 0xf800) != (lcGetAbs() & 0xf800))) {
      if (pass) message(MSG_WARNING, "call_near: long call needed, but near forced. Code will be wrong.");
    }
    mmWrite(chipmem, lcGetAbs(), 0x2000 | (dst & 0x07ff));
  }
  bsr = -1;	// We can't be sure BSR didn't change during call
  lcInc();
}


void
do_call_far(const char *op)
{
  int dst;

  dst = symResolveAddr(op);
  if (!pass && (dst == -1)) {
    fprintf(stderr, "%04x NOTICE: Foward label; assuming/forcing long call\n", lc);
    ocitSetFlag(lcGet(), curpsect->pnum, OC_FORWARD);
  }

  if (pass && (dst == -1)) {
    message(MSG_ERROR, "Unresolved symbol: %s", op);
    exit(1);
  }

  bsr = -1;	// We can't be sure BSR didn't change during call
  message(MSG_INFO, "call_far: long call forced.");
  // Write PCLATH
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x3180 | (dst & 0xff00) >> 8);
  lcInc();
  if (pass && mmHasData(chipmem, lcGetAbs()))
      message(MSG_WARNING, "Overwriting previously written location");
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x2000 | (dst & 0x07ff));
  lcInc();
  // Restore PCLATH
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x3180 | (lcGetAbs() & 0xff00) >> 8);
  lcInc();
}


void
do_call(const char *op)
{
  int dst;
  int far;	/* Is this a far call? */

  if (memmodel == MM_SMALL) {
    do_call_near(op);
    return;
  }

  dst = symResolveAddr(op);
  if (dst == EVAL_FAIL) {
    if (!pass) {
      message(MSG_NOTICE, "Foward label; assuming/forcing long call");
      ocitSetFlag(lcGet(), curpsect->pnum, OC_FORWARD);
    }

    if (pass) {
      message(MSG_ERROR, "Unresolved symbol: %s", op);
      exit(1);
    }
  }

  bsr = -1;	// We can't be sure BSR didn't change during call
  if (!pass) {
    /* Pass 1 */
    if ((dst == EVAL_FAIL) || ocitCheckFlag(lcGet(), OC_FORWARD) || ((dst & 0xf800) != (lcGet() & 0xf800))) {
      message(MSG_INFO, "call: long call to 0x%04x.", dst);
      lcInc();
      lcInc();
    }
    lcInc();
  } else {
    /* Pass 2 */

    if (pass && (dst == EVAL_FAIL)) {
      message(MSG_ERROR, "Unresolved symbol: %s", op);
      exit(1);
    }

    far = ocitCheckFlag(lcGet(), OC_FORWARD) || ((dst & 0xf800) != (lcGetAbs() & 0xf800));
    if (far) {
      message(MSG_INFO, "INFO: call: long call to 0x%04x.", dst);
      if (mmHasData(chipmem, lcGetAbs()))
          message(MSG_WARNING, "Overwriting previously written location (1)");
      mmWrite(chipmem, lcGetAbs(), 0x3180 | (dst & 0xff00) >> 8);
      lcInc();
    }

    if (mmHasData(chipmem, lcGetAbs()))
        message(MSG_WARNING, "Overwriting previously written location (2)");
    mmWrite(chipmem, lcGetAbs(), 0x2000 | (dst & 0x07ff));
    lcInc();

    if (far) {
      if (mmHasData(chipmem, lcGetAbs()))
          message(MSG_WARNING, "Overwriting previously written location (3)");
      mmWrite(chipmem, lcGetAbs(), 0x3180 | (lcGetAbs() & 0xff00) >> 8);
      lcInc();
    }
  }
}


void
do_callw(const char *op)
{
  if (!(proc.inset & IS_EXTENDED))
      MESSAGE(MSG_WARNING, "Invalid opcode for this chip");

  bsr = -1;	// We can't be sure BSR didn't change during call
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x000a);
  lcInc();
}


/*
 * Goto does not need bsr invalidation, since the only
 * way to continue execution past it is via another goto,
 * and the needed label will invalidate it.  But would it
 * be extra security to do it anyway?
 */

void
do_goto_near(const char *op)
{
  int dst;

  dst = symResolveAddr(op);
  if (dst == EVAL_FAIL) {
    if (!pass) {
      message((memmodel == MM_SMALL) ? MSG_INFO : MSG_NOTICE,
          "Forward label; near goto forced");
      ocitSetFlag(lcGet(), curpsect->pnum, OC_FORWARD);
    }
    if (pass) {
      message(MSG_ERROR, "Unresolved symbol: %s", op);
      exit(1);
    }
  }

  if (pass) {
    if ((dst == EVAL_FAIL) || ((dst & 0xf800) != (lcGetAbs() & 0xf800))) {
      message(MSG_WARNING, "goto_near: long goto required; code will be wrong.");
    }
    mmWrite(chipmem, lcGetAbs(), 0x2800 | (dst & 0x07ff));
  }
  lcInc();
}


void
do_goto_far(const char *op)
{
  int dst;

  dst = symResolveAddr(op);
  if (dst == EVAL_FAIL) {
    if (!pass) {
      message(MSG_INFO, "Foward label; long call forced");
      ocitSetFlag(lcGet(), curpsect->pnum, OC_FORWARD);
    }

    if (pass) {
      message(MSG_ERROR, "Unresolved symbol: %s", op);
      exit(1);
    }
  }

  message(MSG_INFO, "goto_far: long goto forced.");
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x3180 | (dst & 0xff00) >> 8);
  lcInc();
  if (pass && mmHasData(chipmem, lcGetAbs()))
      message(MSG_WARNING, "Overwriting previously written location");
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x2800 | (dst & 0x07ff));
  lcInc();
}


void
do_goto(const char *op)
{
  int dst;

  if (memmodel == MM_SMALL) {
    do_goto_near(op);
    return;
  }

  dst = symResolveAddr(op);
  if (dst == EVAL_FAIL) {
    if (!pass) {
      message(MSG_NOTICE, "Forward label; assuming/forcing long goto");
      ocitSetFlag(lcGet(), curpsect->pnum, OC_FORWARD);
    }

    if (pass) {
      message(MSG_ERROR, "Unresolved symbol: %s", op);
      exit(1);
    }
  }

  if (!pass) {
    /* Pass 1 */
    if ((dst == -1) || ocitCheckFlag(lcGet(), OC_FORWARD) || ((dst & 0xf800) != (lcGet() & 0xf800))) {
      message(MSG_INFO, "goto: long goto.");
      lcInc();
    }
  } else {
    /* Pass 2 */
    // TODO: Should we pass lcGetAbs or just lcGet to ocitCheckFlag?
    if ((dst == -1) || ocitCheckFlag(lcGet(), OC_FORWARD) || ((dst & 0xf800) != (lcGetAbs() & 0xf800))) {
      message(MSG_INFO, "goto: long goto.");
      mmWrite(chipmem, lcGetAbs(), 0x3180 | (dst & 0xff00) >> 8);
      lcInc();
      if (pass && mmHasData(chipmem, lcGetAbs()))
          message(MSG_WARNING, "Overwriting previously written location");
    }
    mmWrite(chipmem, lcGetAbs(), 0x2800 | (dst & 0x07ff));
  }
  lcInc();
}


void
do_retfie(const char *op)
{
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x0009);
  lcInc();
}


void
do_retlw(const char *op)
{
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x3400 | (symResolveLiteral(op) & 0x00ff));
  lcInc();
}


void
do_return(const char *op)
{
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x0008);
  lcInc();
}



/***** Inherent Operations *****/

void do_clrwdt(const char *op)
{
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x0064);
  lcInc();
}

void do_nop(const char *op)
{
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x0000);
  lcInc();
}

void do_option(const char *op)
{
  if (!(proc.inset & IS_EXTENDED))
      MESSAGE(MSG_WARNING, "Invalid opcode for this chip");

  message(MSG_WARNING, "OPTION instruction deprecated");
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x0062);
  lcInc();
}

void do_reset(const char *op)
{
  if (!(proc.inset & IS_EXTENDED))
      MESSAGE(MSG_WARNING, "Invalid opcode for this chip");

  if (pass) mmWrite(chipmem, lcGetAbs(), 0x0001);
  lcInc();
}

void do_sleep(const char *op)
{
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x0063);
  lcInc();
}

void do_tris(const char *op)
{
  int f = 0;

  if (!(proc.inset & IS_EXTENDED))
      MESSAGE(MSG_WARNING, "Invalid opcode for this chip");

  message(MSG_WARNING, "TRIS instruction deprecated");
  if ((*op >= '5') && (*op <= '7')) {
    f = *op - '0';
  } else if ((toupper(*op) >= 'A') && (toupper(*op) <= 'C')) {
    f = 5 + (*op - 'A');
  } else {
    fprintf(stderr, "ERROR: Opcode TRIS has invalid operand '%s'\n", op);
    exit(1);
  }
  if (pass) mmWrite(chipmem, lcGetAbs(), 0x0060 | f);
  lcInc();
}


/***** "C Optimized" opcodes *****/

void do_addfsr0(const char *op)
{
  if (!(proc.inset & IS_EXTENDED))
      MESSAGE(MSG_WARNING, "Invalid opcode for this chip");

  if (pass) mmWrite(chipmem, lcGetAbs(), 0x3100 | (shorten2s(symResolveLiteral(op), 6)));
  lcInc();
}


void do_addfsr1(const char *op)
{
  if (!(proc.inset & IS_EXTENDED))
      MESSAGE(MSG_WARNING, "Invalid opcode for this chip");

  if (pass) mmWrite(chipmem, lcGetAbs(), 0x3140 | (shorten2s(symResolveLiteral(op), 6)));
  lcInc();
}


void do_moviw(const char *op)
{
  int m = -1;
  int rp = 0;
  int r;
  int k;
  const char *ch;
  char ofs[80];

  if (!(proc.inset & IS_EXTENDED))
      MESSAGE(MSG_WARNING, "Invalid opcode for this chip");

  if ((ch = strchr(op, '['))) {
    r = (ch[4] - '0') << 6;
    if ((ch - op) > 79) ch = op+79;
    strncpy(ofs, op, (ch - op));
    ofs[(ch - op)] = '\0';
    k = evalString(ofs);
    // sscanf(op, "%d", &k);
    message(MSG_INFO, "r = 0x%04x, k = %d (%s)", r, k, ofs);
    
    if (pass) mmWrite(chipmem, lcGetAbs(), 0x3f00 | r | (shorten2s(k, 6) & 0x003f));
    lcInc();

    return;
  }

  if (strlen(op) < 6) {
    message(MSG_ERROR, "moviw operand syntax error.");
    exit(1);
  }
  if (op[1] == '+') { m = 0; rp = 5; }
  if (op[1] == '-') { m = 1; rp = 5; }
  if (op[4] == '+') { m = 2; rp = 3; }
  if (op[4] == '-') { m = 3; rp = 3; }
  if ((m < 0) || (rp < 3)) {
    message(MSG_ERROR, "moviw operand syntax error.");
    exit(1);
  }

  r = (op[rp] - '0') << 2;

  if (pass) mmWrite(chipmem, lcGetAbs(), 0x0010 | r | m);
  lcInc();
}


void do_movwi(const char *op)
{
  int m = -1;
  int rp = 0;
  int r;
  int k;
  const char *ch;
  char ofs[80];

  if (!(proc.inset & IS_EXTENDED))
      MESSAGE(MSG_WARNING, "Invalid opcode for this chip");

  if ((ch = strchr(op, '['))) {
    r = (ch[4] - '0') << 6;
    if ((ch - op) > 79) ch = op+79;
    strncpy(ofs, op, (ch - op));
    ofs[(ch - op)] = '\0';
    k = evalString(ofs);
    // sscanf(op, "%d", &k);
    
    if (pass) mmWrite(chipmem, lcGetAbs(), 0x3f80 | r | (shorten2s(k, 6) & 0x003f));
    lcInc();

    return;
  }

  if (strlen(op) < 6) {
    message(MSG_ERROR, "movwi operand syntax error.");
    exit(1);
  }
  if (op[1] == '+') { m = 0; rp = 5; }
  if (op[1] == '-') { m = 1; rp = 5; }
  if (op[4] == '+') { m = 2; rp = 3; }
  if (op[4] == '-') { m = 3; rp = 3; }
  if ((m < 0) || (rp < 3)) {
    message(MSG_ERROR, "movwi operand syntax error.");
    exit(1);
  }

  r = (op[rp] - '0') << 2;

  if (pass) mmWrite(chipmem, lcGetAbs(), 0x0018 | r | m);
  lcInc();
}


void do_db(const char *op)
{
  int x;

  x = evalString(op);
  message(MSG_INFO, "do_db: x = %d, curpsect->class = %d", x, curpsect->class);

  if (pass) {
    message(MSG_INFO, "do_db: lcGetAbs() = %d", lcGetAbs());
    switch (curpsect->class) {
      case PS_CODE:
        mmWrite(chipmem, lcGetAbs(), x & 0x00ff);
        break;
      case PS_EEDATA:
        mmWrite(chipmem, lcGetAbs(), x & 0x00ff);
        break;
      case PS_CONFIG:
        mmWrite(chipmem, lcGetAbs(), x & 0x00ff);
        break;
      default:
        break;	/* Not a memory class that gets initialized */
    }
      
  }  /* if pass 2 */
  lcInc();
}


void do_dw(const char *op)
{
  int x;

  x = evalString(op);
  message(MSG_INFO, "do_dw: x = %d, curpsect->class = %d",
      x, curpsect->class);

  if (pass) {
    message(MSG_INFO, "do_dw: lcGetAbs() = %d", lcGetAbs());
    switch (curpsect->class) {
      case PS_CODE:
        mmWrite(chipmem, lcGetAbs(), x & 0x3fff);
        break;
      case PS_EEDATA:
        mmWrite(chipmem, lcGetAbs(), x & 0x00ff);
        mmWrite(chipmem, lcGetAbs()+1, (x & 0xff00) >> 8);
        break;
      case PS_CONFIG:
        mmWrite(chipmem, lcGetAbs(), x & 0x3fff);
        break;
      default:
        break;	/* Not a memory class that gets initialized */
    }
      
  }  /* if pass 2 */
  lcInc();
  if ((curpsect->class != PS_CODE) && (curpsect->class != PS_CONFIG))
      lcInc();
}


void do_ds(const char *op)
{
  int x;
  int i;

  x = evalString(op);
  message(MSG_INFO, "do_ds: x = %d, curpsect->class = %d",
      x, curpsect->class);

  /* We reserve space here, but don't initialize it, even in
   * HEX memory.  So basically, we just advance the location counter
   */
  for (i=0; i<x; i++) lcInc();
}



int do_opcode(void (*opcode_func)(const char *op),
              const char *opcode, const char *operands, const char *comment)
{
  int opstart;
  int i;

  opstart = pass ? lcGetAbs() : lcGet();

  if (psIsHexMem(curpsect->class)) {
    if (pass && mmHasData(chipmem, lcGetAbs()))
        message(MSG_WARNING, "Overwriting previously written location");
  }

  (*opcode_func)(operands);

  // fprintf(stderr, "DEBUG: do_opcode: lc at start = %04x, at end = %04x\n", opstart, lc);

  if (psIsHexMem(curpsect->class)) {
    /* Listing file should only be open during pass 2 */
    if (lstp) {
      fprintf(lstp, "%04x", opstart);
      for (i=opstart; i<lcGetAbs(); i++) fprintf(lstp, " %04hx", mmRead(chipmem, i));
      fprintf(lstp, "\t\t%s %s\t\t%s\n", opcode, operands, comment);
    }
  }  /* if hex mem */

  return ((pass ? lcGetAbs() : lcGet()) - opstart);
}


/*
 *  Checks 'opcode' for a valid opcode. If found, calls do_opcode
 *  and returns its result.  Otherwise returns -1 to indicate
 *  it wasn't a valid opcode (may still be a directive, though).
 */

int try_opcode(const char *opcode, const char *operands,
               const char *comment)
{
  switch (opcode[0]) {
    case 'a':	// 8 opcodes
      if (!strcmp(opcode, "addwf"))
          return do_opcode(do_addwf,  opcode, operands, comment);
      if (!strcmp(opcode, "addwfc"))
          return do_opcode(do_addwfc, opcode, operands, comment);
      if (!strcmp(opcode, "andwf"))
          return do_opcode(do_andwf,  opcode, operands, comment);
      if (!strcmp(opcode, "asrf"))
          return do_opcode(do_asrf,   opcode, operands, comment);
      if (!strcmp(opcode, "addlw"))
          return do_opcode(do_addlw,  opcode, operands, comment);
      if (!strcmp(opcode, "andlw"))
          return do_opcode(do_andlw,  opcode, operands, comment);
      if (!strcmp(opcode, "addfsr0"))
          return do_opcode(do_addfsr0, opcode, operands, comment);
      if (!strcmp(opcode, "addfsr1"))
          return do_opcode(do_addfsr1, opcode, operands, comment);
      break;
    case 'b':	// 6 opcodes
      if (!strcmp(opcode, "bcf"))
          return do_opcode(do_bcf,    opcode, operands, comment);
      if (!strcmp(opcode, "bsf"))
          return do_opcode(do_bsf,    opcode, operands, comment);
      if (!strcmp(opcode, "btfsc"))
          return do_opcode(do_btfsc,  opcode, operands, comment);
      if (!strcmp(opcode, "btfss"))
          return do_opcode(do_btfss,  opcode, operands, comment);
      if (!strcmp(opcode, "bra"))
          return do_opcode(do_bra,    opcode, operands, comment);
      if (!strcmp(opcode, "brw"))
          return do_opcode(do_brw,    opcode, operands, comment);
      break;
    case 'c':	// 8 opcodes
      if (!strcmp(opcode, "clrf"))
          return do_opcode(do_clrf,   opcode, operands, comment);
      if (!strcmp(opcode, "clrw"))
          return do_opcode(do_clrw,   opcode, operands, comment);
      if (!strcmp(opcode, "call"))
          return do_opcode(do_call,   opcode, operands, comment);
      if (!strcmp(opcode, "call_far"))
          return do_opcode(do_call_far,  opcode, operands, comment);
      if (!strcmp(opcode, "call_near"))
          return do_opcode(do_call_near, opcode, operands, comment);
      if (!strcmp(opcode, "comf"))
          return do_opcode(do_comf,   opcode, operands, comment);
      if (!strcmp(opcode, "callw"))
          return do_opcode(do_callw,  opcode, operands, comment);
      if (!strcmp(opcode, "clrwdt"))
          return do_opcode(do_clrwdt, opcode, operands, comment);
      break;
    case 'd':
      if (!strcmp(opcode, "decf"))
          return do_opcode(do_decf,   opcode, operands, comment);
      if (!strcmp(opcode, "decfsz"))
          return do_opcode(do_decfsz, opcode, operands, comment);
      if (!strcmp(opcode, "db"))
          return do_opcode(do_db,     opcode, operands, comment);
      if (!strcmp(opcode, "dw"))
          return do_opcode(do_dw,     opcode, operands, comment);
      if (!strcmp(opcode, "ds"))
          return do_opcode(do_ds,     opcode, operands, comment);
      break;
    case 'e':
    case 'f':
      if (!strcmp(opcode, "fcall"))
          return do_opcode(do_call_far,  opcode, operands, comment);
      break;
    case 'g':
      if (!strcmp(opcode, "goto"))
          return do_opcode(do_goto,   opcode, operands, comment);
      if (!strcmp(opcode, "goto_far"))
          return do_opcode(do_goto_far,  opcode, operands, comment);
      if (!strcmp(opcode, "goto_near"))
          return do_opcode(do_goto_near, opcode, operands, comment);
      break;
    case 'h':
      break;
    case 'i':
      if (!strcmp(opcode, "incf"))
          return do_opcode(do_incf,   opcode, operands, comment);
      if (!strcmp(opcode, "iorwf"))
          return do_opcode(do_iorwf,  opcode, operands, comment);
      if (!strcmp(opcode, "incfsz"))
          return do_opcode(do_incfsz, opcode, operands, comment);
      if (!strcmp(opcode, "iorlw"))
          return do_opcode(do_iorlw,  opcode, operands, comment);
      break;
    case 'j':
    case 'k':
      break;
    case 'l':
      if (!strcmp(opcode, "lgoto"))
          return do_opcode(do_goto_far,  opcode, operands, comment);
      if (!strcmp(opcode, "lslf"))
          return do_opcode(do_lslf,   opcode, operands, comment);
      if (!strcmp(opcode, "lsrf"))
          return do_opcode(do_lsrf,   opcode, operands, comment);
      break;
    case 'm':	// 7 opcodes
      if (!strcmp(opcode, "movf"))
          return do_opcode(do_movf,   opcode, operands, comment);
      if (!strcmp(opcode, "movwf"))
          return do_opcode(do_movwf,  opcode, operands, comment);
      if (!strcmp(opcode, "movlb"))
          return do_opcode(do_movlb,  opcode, operands, comment);
      if (!strcmp(opcode, "movlp"))
          return do_opcode(do_movlp,  opcode, operands, comment);
      if (!strcmp(opcode, "movlw"))
          return do_opcode(do_movlw,  opcode, operands, comment);
      if (!strcmp(opcode, "moviw"))
          return do_opcode(do_moviw,  opcode, operands, comment);
      if (!strcmp(opcode, "movwi"))
          return do_opcode(do_movwi,  opcode, operands, comment);
      break;
    case 'n':
      if (!strcmp(opcode, "nop"))
          return do_opcode(do_nop,    opcode, operands, comment);
      break;
    case 'o':
      if (!strcmp(opcode, "option"))
          return do_opcode(do_option, opcode, operands, comment);
      break;
    case 'p':
    case 'q':
      break;
    case 'r':	// 6 opcodes
      if (!strcmp(opcode, "return"))
          return do_opcode(do_return, opcode, operands, comment);
      if (!strcmp(opcode, "retlw"))
          return do_opcode(do_retlw,  opcode, operands, comment);
      if (!strcmp(opcode, "rlf"))
          return do_opcode(do_rlf,    opcode, operands, comment);
      if (!strcmp(opcode, "rrf"))
          return do_opcode(do_rrf,    opcode, operands, comment);
      if (!strcmp(opcode, "retfie"))
          return do_opcode(do_retfie, opcode, operands, comment);
      if (!strcmp(opcode, "reset"))
          return do_opcode(do_reset,  opcode, operands, comment);
      break;
    case 's':
      if (!strcmp(opcode, "subwf"))
          return do_opcode(do_subwf,  opcode, operands, comment);
      if (!strcmp(opcode, "subwfb"))
          return do_opcode(do_subwfb, opcode, operands, comment);
      if (!strcmp(opcode, "swapf"))
          return do_opcode(do_swapf,  opcode, operands, comment);
      if (!strcmp(opcode, "sublw"))
          return do_opcode(do_sublw,  opcode, operands, comment);
      if (!strcmp(opcode, "sleep"))
          return do_opcode(do_sleep,  opcode, operands, comment);
      break;
    case 't':
      if (!strcmp(opcode, "tris"))
          return do_opcode(do_tris,   opcode, operands, comment);
      break;
    case 'u':
    case 'v':
    case 'w':
      break;
    case 'x':
      if (!strcmp(opcode, "xorwf"))
          return do_opcode(do_xorwf,  opcode, operands, comment);
      if (!strcmp(opcode, "xorlw"))
          return do_opcode(do_xorlw,  opcode, operands, comment);
      break;
    default:
      break;
  }  // switch

  return -1;
}


