/*
 * asm1825 - Assembler for PIC16F1825 and others.
 *
 * $Id: token.c,v 1.1 2021-02-26 13:32:57-05 andrew Exp andrew $
 * 
 * Copyright (c) 2021 Andrew Turnquist
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 * 
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "token.h"
#include "lex.yy.h"
#include "lexerstates.h"

#define TOK_MAX		64

#define TOK_BLANK	0
#define TOK_LABEL	1
#define TOK_ARITH	2
#define TOK_PAREN	3


#define FREE(x) if (x) { free(x); x = NULL; }


const char *lexinput;
int lexinputidx;
int lexinputlim;

#define MIN(a,b)        ((a < b) ? a : b)

int my_yyinput(char *buf, int max_size)
{
  int n = MIN(max_size, lexinputlim - lexinputidx);

  if (n > 0) {
    memcpy(buf, &(lexinput[lexinputidx]), n);
    lexinputidx += n;
  }
  return n;
}


int tokIsLabel(const Token *t)
{
  return (t->tok_type == STRING);
  // if (isalnum(*ch)) return 1;
  // if (*ch == '@') return 1;
  // if (*ch == '_') return 1;
  // if (*ch == '$') return 1;
  // return 0;
}


int tokIsNumber(const Token *t)
{
  if (t->tok_type == INTEGER) return 1;
  if (t->tok_type == HEXNUM)  return 1;
  if (t->tok_type == OCTNUM)  return 1;
  if (t->tok_type == BINNUM)  return 1;
  return 0;
  // if (isdigit(*ch)) return 1;
  // if (ch[1] == '\0') return 0;
  // if ((ch[0] == '-') && (isdigit(ch[1]))) return 1;
  // return 0;
}


int tokIsArith(const char *ch)
{
  if (*ch == '+') return 1;
  if (*ch == '-') return 1;
  if (*ch == '*') return 1;
  if (*ch == '/') return 1;
  if (*ch == '%') return 1;
  if (*ch == '<') return 1;
  if (*ch == '>') return 1;
  if (*ch == '&') return 1;
  if (*ch == '|') return 1;
  if (*ch == '^') return 1;
  return 0;
}


int tokIsParen(const char *ch)
{
  if (*ch == '(') return 1;
  if (*ch == ')') return 1;
  return 0;
}


int tokType(const Token *t)
{
  return (t->tok_type);
  // if (isblank(*ch)) return TOK_BLANK;
  // if (tokIsLabel(ch)) return TOK_LABEL;
  // if (tokIsArith(ch)) return TOK_ARITH;
  // if (tokIsParen(ch)) return TOK_PAREN;
  // return TOK_BLANK;	// Unknown characters treated as blanks
}


Token **tokSplit(const char *str)
{
  int argc = 0;
  Token **argv;
  int toktype;

  argv = (Token **)calloc(TOK_MAX, sizeof(Token *));

  argc = 0;
  lexinput = str;
  lexinputidx = 0;
  lexinputlim = strlen(lexinput);

  while ((argc < TOK_MAX) && (toktype = yylex())) {
    if (toktype) {
      // fprintf(stderr, "Pre  %d: %d -> '%s'\n", argc, toktype, yytext);
      argv[argc] = (Token *)malloc(sizeof(Token));
      argv[argc]->tok_type = toktype;
      argv[argc]->tok_text = strdup(yytext);
      // fprintf(stderr, "Post %d: %d -> '%s'\n", argc, argv[argc]->tok_type, argv[argc]->tok_text);
      argc++;
    } else {
      argv[argc] = NULL;
      break;
    }
  }
  argv[argc] = NULL;

  return argv;
}


void tokFree(Token **tokv)
{
  int i;

  i=0;
  while (tokv[i]) {
    FREE(tokv[i]->tok_text);
    FREE(tokv[i]);
    i++;
  }
  free(tokv);
}


int tokCount(Token ** tokv)
{
  int count = 0;
  while (tokv[count] != NULL) count++;
  return count;
}


// Just to wrap my head around this...
//   tokv is a pointer to an array of character pointers (strings).
//   What I need: passed tokv itself should not change, but
//       the list of character pointers it points to should change
//       and those changes should be returned.
void tokShiftL(Token ** tokv, int pos, int count)
{
  int j;

  j = 0;
  do {
    // fprintf(stderr, "    tokShiftL: tokv[%d] %s", pos+j, tokv[pos+j]);
    if (j<count) FREE(tokv[pos+j]);
    tokv[pos+j] = tokv[pos+j+count];
    // fprintf(stderr, " <-- %s\n", tokv[pos+j]);
    if (tokv[pos+j] == NULL) break;
    j++;
  } while (1);
}


Token **tokDup(Token **tokv, int start, int end)
{
  int len;
  int i;
  Token **argv;

  len = (end - start) + 2;
  argv = (Token **)calloc(len, sizeof(Token *));
  for (i=start; i<=end; i++) {
    argv[i-start] = (Token *)malloc(sizeof(Token));
    argv[i-start]->tok_type = tokv[i]->tok_type;
    argv[i-start]->tok_text = strdup(tokv[i]->tok_text);
  }
  argv[i-start] = NULL;
  return argv;
}


void tokDump(Token **tokv)
{
  int i = 0;
  while (tokv[i]) {
    printf("%d: %d -> '%s'\n", i, tokv[i]->tok_type, tokv[i]->tok_text);
    i++;
  }
}

